import bpy
from bpy.app.handlers import persistent

is_transforming = False
was_transforming = False
booleon_cutter_info_dict = {}
temp_cutter_clones = []

def reset_globals():
    global is_transforming, was_transforming, booleon_cutter_info_dict, temp_cutter_clones
    is_transforming = False
    was_transforming = False
    booleon_cutter_info_dict = {}
    temp_cutter_clones = []

def create_copy_of_cutter_object(cutting_object):
    c = bpy.context
    depsgraph = c.evaluated_depsgraph_get()
    cutting_object_eval = cutting_object.evaluated_get(depsgraph)
    new_cutter = cutting_object_eval.copy()
    new_cutter.data = cutting_object_eval.data.copy()
    c.collection.objects.link(new_cutter)
 
    return new_cutter

def find_boolean_relationships():
    c = bpy.context
    selected_objects = set(c.selected_objects)  # Use a set for O(1) lookups, also de-duplicates duplicates (not that it matters in this case, but I want to remember to it for the future)
    if not selected_objects:
        return

    boolean_links = {}

    objects_with_modifiers = (obj for obj in c.scene.objects if obj.modifiers)

    for obj in objects_with_modifiers:
        for mod in obj.modifiers:
            # Check if the modifier is a Boolean modifier and references a selected object
            if mod.type == 'BOOLEAN' and mod.object in selected_objects:
                cutting_object = mod.object
                boolean_links.setdefault(cutting_object, []).append({
                    'target_object': obj,
                    'modifier': mod
                })

    return boolean_links

    # Print results
    # for cutting_object, links in boolean_links.items():
    #     print(f"Boolean Object: {cutting_object.name}")
    #     for link in links:
    #         print(f"  Cuts Into: {link['target_object']}, Modifier: {link['modifier']}")


def faster_objs_delete(objs): # deleting objects in Blender is crazy slow by deafult, this is a faster way to do it
    for obj in objs:
        for collection in obj.users_collection:
            collection.objects.unlink(obj)

def is_current_running_modal_transform_none_live_booleans(dummy1, dummy2):
    global is_transforming, was_transforming, booleon_cutter_info_dict, temp_cutter_clones
    c = bpy.context
    is_transforming = False
    
    if c.mode == 'OBJECT':
        selected = c.selected_objects
        if selected:
            wm = c.window_manager

            for window in wm.windows:
                for op in window.modal_operators:
                    op_prefix = op.bl_idname.split("_OT_")
            
                    if "TRANSFORM" in op_prefix:
                        is_transforming = True
                        break

            if is_transforming:
                if not booleon_cutter_info_dict:
                    booleon_cutter_info_dict = find_boolean_relationships()
                
                    for cutting_object, links in booleon_cutter_info_dict.items():
                        for link in links:
                            if link['modifier'].object:
                                link['modifier'].object = None

                # selected_objects_names = [obj.name for obj in selected]
                # print("Transforming: " + str(selected_objects_names))
            elif was_transforming and not is_transforming:
                # print("Stopped transforming")

                if not temp_cutter_clones:
                    for cutting_object, links in booleon_cutter_info_dict.items():
                        for link in links:
                            if link['modifier'].object != cutting_object:
                                link['modifier'].object = cutting_object
                else:
                    iteration = -1
                    for cutting_object, links in booleon_cutter_info_dict.items():
                        iteration += 1
                        for link in links:
                            cutting_object.hide_set(False)
                            cutting_object.select_set(True)

                            # copy the location, rotation and scale of the cutter object to the boolean object
                            cutting_object.location = temp_cutter_clones[iteration].location
                            cutting_object.rotation_euler = temp_cutter_clones[iteration].rotation_euler
                            cutting_object.scale = temp_cutter_clones[iteration].scale
                    # remove the temp cutter clones, too slow in bigger scenes, just let Blender handle the unused data instead
                    # for new_cutter in temp_cutter_clones:
                        # get new_cutter mesh data from the object
                        # new_cutter_mesh = new_cutter.data
                        # bpy.data.objects.remove(new_cutter)
                        # bpy.data.meshes.remove(new_cutter_mesh)
                    faster_objs_delete(temp_cutter_clones)

                booleon_cutter_info_dict = {}
                temp_cutter_clones = []

            was_transforming = is_transforming


# We need to detect if the user presses G, R, S, to invoke a transform operator, to skip any initial lag spikes due to recalculations of the modifiers.
# Seems like the only way to get over the first lag spike when transforming. RIP Gizmo users, I guess.
class ENTERING_TRANSFORM_OT_None_Live_Booleans(bpy.types.Operator):
    bl_idname = "transform.entering_transform_none_live_booleans"
    bl_label = "Entering Transform"
    bl_options = {'INTERNAL'}

    def execute(self, context):
        global booleon_cutter_info_dict, temp_cutter_clones
        booleon_cutter_info_dict = find_boolean_relationships()
        
        for cutting_object, links in booleon_cutter_info_dict.items():
            for link in links:
                # hide the cutter object
                cutting_object.hide_set(True)
                cutting_object.select_set(False)
                
                # create a copy of the cutter object
                new_cutter = create_copy_of_cutter_object(cutting_object)
                new_cutter.select_set(True)
                temp_cutter_clones.append(new_cutter)
        return {'PASS_THROUGH'}

    def register():
        # add the keymap
        wm = bpy.context.window_manager
        km = wm.keyconfigs.addon.keymaps.new(name='Object Mode', space_type='EMPTY')
        kmi = km.keymap_items.new(ENTERING_TRANSFORM_OT_None_Live_Booleans.bl_idname, type='G', value='PRESS', shift=False, ctrl=False, alt=False, oskey=False)
        kmi = km.keymap_items.new(ENTERING_TRANSFORM_OT_None_Live_Booleans.bl_idname, type='R', value='PRESS', shift=False, ctrl=False, alt=False, oskey=False)
        kmi = km.keymap_items.new(ENTERING_TRANSFORM_OT_None_Live_Booleans.bl_idname, type='S', value='PRESS', shift=False, ctrl=False, alt=False, oskey=False)

        bpy.app.handlers.load_post.append(unregister_none_live_booleans)

    def unregister():
        if is_current_running_modal_transform_none_live_booleans in bpy.app.handlers.depsgraph_update_pre:
            bpy.app.handlers.depsgraph_update_pre.remove(is_current_running_modal_transform_none_live_booleans)
     
        if unregister_none_live_booleans in bpy.app.handlers.load_post:
            bpy.app.handlers.load_post.remove(unregister_none_live_booleans)

        # remove the keymap
        wm = bpy.context.window_manager
        km = wm.keyconfigs.addon.keymaps['Object Mode']

        for kmi in km.keymap_items:
            if kmi.idname == ENTERING_TRANSFORM_OT_None_Live_Booleans.bl_idname:
                km.keymap_items.remove(kmi)

        reset_globals()
    

@persistent
def unregister_none_live_booleans(dummy, dummy2):
    try:
        bpy.utils.unregister_class(ENTERING_TRANSFORM_OT_None_Live_Booleans)
    except:
        pass
